@@ -38,6 +38,8 @@
 use core::borrow::Borrow;
 use core::cmp::Ordering;
 use core::fmt;
+use core::num::IntErrorKind;
+use core::str::FromStr;
 #[cfg(feature = "std")]
 use std::error::Error;
 
@@ -52,6 +54,43 @@ impl fmt::Display for TryFromIntError {
 #[cfg(feature = "std")]
 impl Error for TryFromIntError {}
 
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct ParseIntError {
+    kind: IntErrorKind,
+}
+
+impl ParseIntError {
+    /// Outputs the detailed cause of parsing an integer failing.
+    #[allow(clippy::missing_const_for_fn)] // This function is not const because the counterpart of stdlib isn't
+    pub fn kind(&self) -> &IntErrorKind {
+        &self.kind
+    }
+
+    // Copied from unstable `core::fmt::ParseIntError::__description` in order to implement
+    // `fmt::Display`.  Once the function is stabilized, this function can be removed.
+    #[doc(hidden)]
+    #[allow(clippy::missing_const_for_fn)] // This function is not const because the counterpart of stdlib isn't
+    pub fn __description(&self) -> &str {
+        match self.kind {
+            IntErrorKind::Empty => "cannot parse integer from empty string",
+            IntErrorKind::InvalidDigit => "invalid digit found in string",
+            IntErrorKind::PosOverflow => "number too large to fit in target type",
+            IntErrorKind::NegOverflow => "number too small to fit in target type",
+            IntErrorKind::Zero => "number would be zero for non-zero type",
+            _ => "Unknown Int error kind",
+        }
+    }
+}
+
+impl fmt::Display for ParseIntError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        self.__description().fmt(f)
+    }
+}
+
+#[cfg(feature = "std")]
+impl Error for ParseIntError {}
+
 macro_rules! const_try_opt {
     ($e:expr) => {
         match $e {
@@ -410,6 +449,22 @@ macro_rules! impl_ranged {
             }
         })*
 
+        impl<const MIN: $internal, const MAX: $internal> FromStr for $type<MIN, MAX> {
+            type Err = ParseIntError;
+            fn from_str(s: &str) -> Result<Self, Self::Err> {
+                let value = s.parse::<$internal>().map_err(|e| ParseIntError {
+                    kind: e.kind().clone()
+                })?;
+                if value < MIN {
+                    Err(ParseIntError { kind: IntErrorKind::NegOverflow })
+                } else if value > MAX {
+                    Err(ParseIntError { kind: IntErrorKind::PosOverflow })
+                } else {
+                    Ok(Self(value))
+                }
+            }
+        }
+
         #[cfg(feature = "serde")]
         impl<const MIN: $internal, const MAX: $internal> serde::Serialize for $type<MIN, MAX> {
             fn serialize<S: serde::Serializer>(&self, serializer: S) -> Result<S::Ok, S::Error> {
@@ -558,4 +613,17 @@ mod tests {
             Err(TryFromIntError)
         );
     }
+
+    #[test]
+    #[allow(clippy::enum_glob_use)]
+    fn test_from_str() {
+        type Target = U32<100, 200>;
+        use std::num::IntErrorKind as Kind;
+        assert!(matches!("50".parse::<Target>(), Err(e) if e.kind() == &Kind::NegOverflow));
+        assert!(matches!("100".parse::<Target>(), Ok(U32(100))));
+        assert!(matches!("150".parse::<Target>(), Ok(U32(150))));
+        assert!(matches!("200".parse::<Target>(), Ok(U32(200))));
+        assert!(matches!("250".parse::<Target>(), Err(e) if e.kind() == &Kind::PosOverflow));
+        assert!(matches!("abc".parse::<Target>(), Err(e) if e.kind() == &Kind::InvalidDigit));
+    }
 }
